package org.openxma.dsl.core.validation;

import static com.google.common.collect.Iterators.filter;
import static org.eclipse.emf.ecore.util.EcoreUtil.getRootContainer;
import static org.eclipse.xtext.util.SimpleAttributeResolver.NAME_RESOLVER;

import java.util.Collection;
import java.util.Iterator;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.xtext.util.Strings;
import org.eclipse.xtext.validation.Check;
import org.eclipse.xtext.validation.CheckType;
import org.openxma.dsl.core.CorePackage;
import org.openxma.dsl.core.model.BooleanParameterValue;
import org.openxma.dsl.core.model.DataTypeAndTypeParameter;
import org.openxma.dsl.core.model.Import;
import org.openxma.dsl.core.model.IntegerParameterValue;
import org.openxma.dsl.core.model.Model;
import org.openxma.dsl.core.model.ModelElement;
import org.openxma.dsl.core.model.ParameterDefinition;
import org.openxma.dsl.core.model.ParameterDefinitionType;
import org.openxma.dsl.core.model.ParameterValue;
import org.openxma.dsl.core.model.SimpleType;
import org.openxma.dsl.core.model.StringParameterValue;
import org.openxma.dsl.core.model.Type;
import org.openxma.dsl.core.model.Validator;
import org.openxma.dsl.core.model.ValidatorReference;

import com.google.common.base.Function;
import com.google.common.base.Predicate;
import com.google.common.collect.Lists;

public class CoreDslJavaValidator extends AbstractCoreDslJavaValidator {
	public static final String MISSING_NAMESPACE = "DOM_MISSING_NAMESPACE";
	public static final String DEPRECATED_IMPORT_URI = "DEPRECATED_IMPORT_URI";
	
	@Check(CheckType.NORMAL)
	public void checkImport(Import _import) {
		if (_import.getImportURI()!=null && !getContext().containsKey(_import)) {
			getContext().put(_import, _import);
			warning("Deprecated Import URI.", CorePackage.IMPORT__IMPORT_URI, DEPRECATED_IMPORT_URI,_import.getImportURI());
		}
	}
	
	@Check(CheckType.NORMAL)
	public void checkNamespace(ModelElement modelElement) {
		EObject container = modelElement.eContainer();
		if (container instanceof Model) {
			Model model = (Model) container;
			if (!getContext().containsKey(modelElement.eContainer()) && !model.eIsSet(CorePackage.eINSTANCE.getModel_Name()) && !Strings.isEmpty(model.getName())) {
				getContext().put(modelElement.eContainer(), modelElement.eContainer());
				warning("Missing Namespace declaration.", CorePackage.MODEL, MISSING_NAMESPACE,model.getName());
			}
		}
	}

	@Check(CheckType.NORMAL)
	public void checkDataTypeUnique(final Type dataType) {
		final String dataTypeId = NAME_RESOLVER.apply(dataType);
		Iterator<EObject> dataTypes = filter(getRootContainer(dataType).eAllContents(), new Predicate<EObject>() {
			public boolean apply(EObject eObject) {
				return eObject instanceof Type && !dataType.equals(eObject)
						&& NAME_RESOLVER.apply(eObject).equals(dataTypeId);
			}
		});
		if (dataTypes.hasNext()) {
			error("Duplicate type '" + dataTypeId + "' detected", null);
		}
	}

	@Check(CheckType.NORMAL)
	public void checkValidatorUnique(final Validator validator) {
		final String validatorId = NAME_RESOLVER.apply(validator);
		Iterator<EObject> validators = filter(getRootContainer(validator).eAllContents(), new Predicate<EObject>() {
			public boolean apply(EObject eObject) {
				return eObject instanceof Validator && !validator.equals(eObject)
						&& NAME_RESOLVER.apply(eObject).equals(validatorId);
			}
		});
		if (validators.hasNext()) {
			error("Duplicate validator '" + validatorId + "' detected", null);
		}
	}

	@Check(CheckType.NORMAL)
	public void checkTypeDef(SimpleType simpleType) {
		Collection<String> validatorParameterNames = Lists.transform(simpleType.getValidatorReference().getParameterValues(),
				new Function<ParameterValue, String>() {
					public String apply(ParameterValue parameterValue) {
						return NAME_RESOLVER.apply(parameterValue.getReferencedParameter());
					}
				});
		Collection<String> typeParameterNames = Lists.transform(simpleType.getTypeParameter(), NAME_RESOLVER);
		if (!validatorParameterNames.containsAll(typeParameterNames)) {
			error("Parameter reference for validator not defined in type '" + NAME_RESOLVER.apply(simpleType) + "'",
					CorePackage.SIMPLE_TYPE__TYPE_PARAMETER);
		}
	}
	
	@Check(CheckType.NORMAL)		
	public void checkParameterCountIsValid(DataTypeAndTypeParameter typeAndParameter) {
		Type type = typeAndParameter.getDataType();
		if (type instanceof SimpleType) {
			SimpleType simpleType = (SimpleType)type;
			EList<ParameterDefinition> parms =simpleType.getTypeParameter();
			int parDefSize = 0;			
			if (parms != null) {
				parDefSize = parms.size();				
			}
			EList<ParameterValue> paramValues = typeAndParameter.getParameterValues();
			int parValSize = 0;
			if (paramValues != null) {
				parValSize = paramValues.size();
			}
			if (parValSize > parDefSize) {
				error("Not more then "+parDefSize+" parameters are accepted (currently "+parValSize+" given	)", CorePackage.DATA_TYPE_AND_TYPE_PARAMETER);
			}
		}		
	}
	
	@Check(CheckType.NORMAL)		
	public void checkParameterTypeMatch(DataTypeAndTypeParameter typeAndParameter) {
		Type type = typeAndParameter.getDataType();
		if (type instanceof SimpleType) {
			SimpleType simpleType = (SimpleType)type;
			EList<ParameterDefinition> parms =simpleType.getTypeParameter();
			int parDefSize = 0;			
			if (parms != null) {
				parDefSize = parms.size();				
			}
			EList<ParameterValue> paramValues = typeAndParameter.getParameterValues();
			int parValSize = 0;
			if (paramValues != null) {
				parValSize = paramValues.size();
			}
			int minSize = Math.min(parDefSize,parValSize);
			doTypeMatchCheck(parms, paramValues, minSize);
		}		
	}
	
	@Check(CheckType.NORMAL)		
	public void checkValidatorParameterCountIsValid(ValidatorReference validatorReference) {		
		EList<ParameterDefinition> parms = validatorReference.getValidator().getValidatorParameter();
		int parDefSize = 0;			
		if (parms != null) {
			parDefSize = parms.size();				
		}
		EList<ParameterValue> paramValues = validatorReference.getParameterValues();
		int parValSize = 0;
		if (paramValues != null) {
			parValSize = paramValues.size();
		}
		if (parValSize > parDefSize) {
			error("Not more then "+parDefSize+" validator parameters are accepted (currently "+parValSize+" given	)", CorePackage.DATA_TYPE_AND_TYPE_PARAMETER);
		}			
	}
	
	@Check(CheckType.NORMAL)		
	public void checkValidatorParameterTypeMatch(ValidatorReference validatorReference) {		
		EList<ParameterDefinition> parms = validatorReference.getValidator().getValidatorParameter();
		int parDefSize = 0;			
		if (parms != null) {
			parDefSize = parms.size();				
		}
		EList<ParameterValue> paramValues = validatorReference.getParameterValues();
		int parValSize = 0;
		if (paramValues != null) {
			parValSize = paramValues.size();
		}
		int minSize = Math.min(parDefSize,parValSize);
		doTypeMatchCheck(parms, paramValues, minSize);		
	}
	
	private void doTypeMatchCheck(EList<ParameterDefinition> parms,
			EList<ParameterValue> paramValues, int minSize) {
		for (int i=0;i<minSize;i++) {
			ParameterDefinition def = parms.get(i);
			ParameterValue val = paramValues.get(i);
			if (val.getReferencedParameter() != null) {
				ParameterDefinition refParDef = val.getReferencedParameter();
				if (refParDef.getType().getValue() != def.getType().getValue()) {
					error("The type of the referenced parameter does not match the expected parameter type",CorePackage.DATA_TYPE_AND_TYPE_PARAMETER);
				}				
			} else {
				switch(def.getType().getValue()) {
				case ParameterDefinitionType.INTEGER_VALUE:
					if (val instanceof StringParameterValue || val instanceof BooleanParameterValue) {
						error("Integer parameter expected", CorePackage.DATA_TYPE_AND_TYPE_PARAMETER);
					}
					break;
				case ParameterDefinitionType.STRING_VALUE:
					if (val instanceof IntegerParameterValue || val instanceof BooleanParameterValue) {
						error("String parameter expected", CorePackage.DATA_TYPE_AND_TYPE_PARAMETER);
					}
					break;
				case ParameterDefinitionType.BOOLEAN_VALUE:
					if (val instanceof IntegerParameterValue || val instanceof StringParameterValue) {
						error("Boolean parameter expected", CorePackage.DATA_TYPE_AND_TYPE_PARAMETER);
					}					
					break;
				default:
				}
			}
		}
	}		

}
